/****************************************************************************
**
** Copyright (C) 2006-2009 fullmetalcoder <fullmetalcoder@hotmail.fr>
**
** This file is part of the Edyuk project <http://edyuk.org>
** 
** This file may be used under the terms of the GNU General Public License
** version 3 as published by the Free Software Foundation and appearing in the
** file GPL.txt included in the packaging of this file.
**
** This file is provided AS IS with NO WARRANTY OF ANY KIND, INCLUDING THE
** WARRANTY OF DESIGN, MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE.
**
****************************************************************************/

#include "qdocumentbuffer.h"

// just for the line endings enum...
#include "qdocument.h"

#include <QFile>
#include <QTextCodec>
#include <QTextDecoder>

/*
	Notes on design :
	
	The idea is to fragment the storage to workaround the issue of Qt
	containers alocation model (too much memory usage when number of items
	grow) and provide faster modification of the content.
	
	The number one goal is to keep lookup by index as fast possible to avoid
	significant impact on document iteration/drawing speed
	
	for such a storage to be useful the block size may not be fixed but instead
	must be kept around an "average" value.
*/

QDocumentBuffer::QDocumentBuffer(QDocument *d)
 :
	m_safetyRoom(10), m_optimalSize(5 * 1024), m_forkThresold(7 * 1024), m_mergeThresold(100),
	m_count(0), m_doc(d),
	m_lastBlock(-1),
	m_originalLineEnding(QDocument::Local)
{
	
}

QDocumentBuffer::~QDocumentBuffer()
{
	clear();
}

void QDocumentBuffer::setContent(const QString& text)
{
	// TODO : smart clear?
	clear();
	
	m_blocks << new Block(0);
	
	int _lf = 0, _cr = 0, _crlf = 0;
	int index = 0, lastNewLine = 0, max = text.count();
	
	const quint16 *uni = text.utf16();
	const QChar *d = text.constData();
	
	while ( index < max )
	{
		quint16 uc = uni[index];
		
		if ( uc == '\n' )
		{
			appendLine(new QDocumentLineHandle(QString(d + lastNewLine, index - lastNewLine)));
			lastNewLine = ++index;
			++_lf;
		} else if ( uc == '\r' ) {
			appendLine(new QDocumentLineHandle(QString(d + lastNewLine, index - lastNewLine)));
			uc = uni[++index];
			
			if ( uc == '\n' ) //c == lf )
			{
				lastNewLine = ++index;
				++_crlf;
			} else {
				lastNewLine = index;
				++_cr;
			}
		} else {
			++index;
		}
	}
	
	const bool dosVSnix = _crlf > _lf;
	const bool dosVSmac = _crlf > _cr;
	const bool nixVSmac = _lf > _cr;
	
	// nothing loaded thus far
	if ( dosVSnix || dosVSmac || nixVSmac )
		m_originalLineEnding = dosVSnix
								?
									(dosVSmac ? QDocument::Windows : QDocument::OldMac)
								:
									(nixVSmac ? QDocument::Unix : QDocument::OldMac);
	else
		m_originalLineEnding = QDocument::Local;
	
	if ( !m_count )
		appendLine(new QDocumentLineHandle());
}

struct LineEndingsCounter
{
	inline LineEndingsCounter() : _cr(0), _lf(0), _crlf(0) {}
	
	int _cr, _lf, _crlf;
};

static QHash<void*, QString> m_chunkLeftovers;
static QHash<void*, LineEndingsCounter> m_chunkCounters;

void QDocumentBuffer::startChunkLoading()
{
	clear();
	
	m_blocks << new Block(0);
	
	m_chunkCounters[this];
}

void QDocumentBuffer::stopChunkLoading()
{
	QString leftover = m_chunkLeftovers[this];
	
	if ( leftover.count() )
		appendLine(new QDocumentLineHandle(leftover));
	
	const LineEndingsCounter& c = m_chunkCounters[this];
	
	const bool dosVSnix = c._crlf > c._lf;
	const bool dosVSmac = c._crlf > c._cr;
	const bool nixVSmac = c._lf > c._cr;
	
	m_chunkCounters.remove(this);
	
	// nothing loaded thus far
	if ( dosVSnix || dosVSmac || nixVSmac )
		m_originalLineEnding = dosVSnix
								?
									(dosVSmac ? QDocument::Windows : QDocument::OldMac)
								:
									(nixVSmac ? QDocument::Unix : QDocument::OldMac);
	else
		m_originalLineEnding = QDocument::Local;
	
}

void QDocumentBuffer::addChunk(const QString& chunk)
{
	LineEndingsCounter& c = m_chunkCounters[this];
	
	QString leftover = m_chunkLeftovers[this];
	leftover += chunk;
	
	const quint16 *uni = leftover.utf16();
	const QChar *d = leftover.constData();
	int index = 0, lastNewLine = 0, max = leftover.count();
	
	while ( index < max )
	{
		quint16 uc = uni[index];
		
		if ( uc == '\n' )
		{
			appendLine(new QDocumentLineHandle(QString(d + lastNewLine, index - lastNewLine)));
			lastNewLine = ++index;
			++c._lf;
		} else if ( uc == '\r' ) {
			appendLine(new QDocumentLineHandle(QString(d + lastNewLine, index - lastNewLine)));
			uc = uni[++index];
			
			if ( uc == '\n' ) //c == lf )
			{
				lastNewLine = ++index;
				++c._crlf;
			} else {
				lastNewLine = index;
				++c._cr;
			}
		} else {
			++index;
		}
	}
	
	m_chunkLeftovers[this] = leftover.mid(lastNewLine);
}

void QDocumentBuffer::load(const QString& filename, QTextCodec *c)
{
	static const QLatin1Char lf('\n'), cr('\r');
	
	// TODO : smart clear?
	clear();
	
	m_blocks << new Block(0);
	
	QFile file(filename);
	int _lf = 0, _cr = 0, _crlf = 0;
	
	forever
	{
		if ( !file.open(QFile::ReadOnly) )
		{
			qWarning("opening failed.");
			break;
		}
		
		// TODO : use memory mapping when available?
		qint64 sz = file.size();
		const qint64 maxBlockSize = 256 * 1024;
		
		QString text;
		QTextDecoder *d = c ? c->makeDecoder() : 0;
		QByteArray buffer(qMin(file.size(), maxBlockSize), 0);
		
		while ( sz > 0 )
		{
			int c = file.read(buffer.data(), buffer.size());
			
			if ( c == -1 )
			{
				qWarning("loading error.");
				break;
			}
			
			if ( !c )
			{
				// eof reached
				break;
			}
			
			if ( d )
				text.append(d->toUnicode(buffer.constData(), c));
			else
				text.append(QString::fromLocal8Bit(buffer));
			
			
			int index = 0, lastNewLine = 0, max = text.count();
			
			const quint16 *uni = text.utf16();
			const QChar *d = text.constData();
			
			while ( index < max )
			{
				//QChar c = d[index]; //text.at(index);
				
				quint16 uc = uni[index];
				// = text.mid(lastNewLine, index - lastNewLine);
				if ( uc == '\n' )
				{
					appendLine(new QDocumentLineHandle(QString(d + lastNewLine, index - lastNewLine)));
					lastNewLine = ++index;
					++_lf;
				} else if ( uc == '\r' ) {
					appendLine(new QDocumentLineHandle(QString(d + lastNewLine, index - lastNewLine)));
					//c = text.at(++index);
					uc = uni[++index];
					
					if ( uc == '\n' )
					{
						lastNewLine = ++index;
						++_crlf;
					} else {
						lastNewLine = index;
						++_cr;
					}
				} else {
					++index;
				}
			}
			
			if ( lastNewLine != -1 )
			{
				text.remove(0, lastNewLine);
			}
		}
		
		if ( text.count() )
			appendLine(new QDocumentLineHandle(text));
		
		break;
	}
	
	const bool dosVSnix = _crlf > _lf;
	const bool dosVSmac = _crlf > _cr;
	const bool nixVSmac = _lf > _cr;
	
	//qDebug("%i, %i, %i", _crlf, _lf, _cr);
	
	// nothing loaded thus far
	if ( dosVSnix || dosVSmac || nixVSmac )
		m_originalLineEnding = dosVSnix
								?
									(dosVSmac ? QDocument::Windows : QDocument::OldMac)
								:
									(nixVSmac ? QDocument::Unix : QDocument::OldMac);
	else
		m_originalLineEnding = QDocument::Local;
	
	//qDebug("=> %i", m_originalLineEnding);
	
	if ( !m_count )
		appendLine(new QDocumentLineHandle());
}

int QDocumentBuffer::count() const
{
	return m_count;
}

int QDocumentBuffer::originalLineEnding() const
{
	return m_originalLineEnding;
}

QDocumentBuffer::iterator QDocumentBuffer::begin() const
{
	return iterator(this, 0, 0);
}

QDocumentBuffer::iterator QDocumentBuffer::end() const
{
	return iterator(this, m_blocks.count() - 1, m_count - 1);
}

QDocumentBuffer::iterator QDocumentBuffer::find(int index) const
{
	return iterator(this, blockForLine(index), index);
}

void QDocumentBuffer::clear()
{
	qDeleteAll(m_blocks);
	m_blocks.clear();
	m_lastBlock = -1;
	m_count = 0;
}

int QDocumentBuffer::blockForLine(int index) const
{
	if ( index < 0 || index >= m_count )
		return -1;
	
	Block *b = 0;
	int low = 0, high = m_blocks.count() - 1;
	
	if ( m_lastBlock != -1 )
	{
		// tries to smartly improve perf for consecutive search
		b = m_blocks.at(m_lastBlock);
		
		if ( index < b->start )
			high = qMax(low, m_lastBlock - 1);
		else if ( b->end <= index )
			low = qMin(high, m_lastBlock + 1);
		else
			return m_lastBlock;
		 
	}
	
	if ( index < m_blocks.at(low)->end )
	{
		m_lastBlock = low;
		return low;
	}
	
	if ( index >= m_blocks.at(high)->start )
	{
		m_lastBlock = high;
		return high;
	}
	
	while ( low < high )
	{
		int mid = (low + high) / 2;
		b = m_blocks.at(mid);
		
		if ( index < b->start )
			high = mid;
		else if ( b->end <= index )
			low = mid;
		else {
			m_lastBlock = mid;
			return mid;
		}
	}
	
	m_lastBlock = low;
	
	return low;
}

QDocumentLineHandle* QDocumentBuffer::first() const
{
	return m_count ? m_blocks.at(0)->lines.first() : 0;
}

QDocumentLineHandle* QDocumentBuffer::last() const
{
	return m_count ? m_blocks.last()->lines.last() : 0;
}

QDocumentLineHandle* QDocumentBuffer::at(int index) const
{
	int blockIndex = blockForLine(index);
	
	return blockIndex != -1 ? m_blocks.at(blockIndex)->at(index) : 0;
}

int QDocumentBuffer::indexOf(const QDocumentLineHandle *h) const
{
	int blockIndex = 0;
	const int count = m_blocks.count();
	
	while ( blockIndex < count )
	{
		Block *b = m_blocks.at(blockIndex);
		
		int idx = b->indexOf(h);
		
		if ( idx != -1 )
			return b->start + idx;
		
		++blockIndex;
	}
	
	return -1;
}

void QDocumentBuffer::fixBlocksFrom(int lastValidBlock)
{
	// update block boundaries
	if ( lastValidBlock < 0 )
		lastValidBlock = 0;
	if ( lastValidBlock >= m_blocks.count() )
		return;
	
	Block *b = 0;
	int previousBoundary = lastValidBlock ? m_blocks.at(lastValidBlock - 1)->end : 0;
	
	while ( lastValidBlock < m_blocks.count() )
	{
		b = m_blocks.at(lastValidBlock);
		b->start = previousBoundary;
		previousBoundary += b->lines.count();
		b->end = previousBoundary;
		
		++lastValidBlock;
	}
}

// for loading
void QDocumentBuffer::appendLine(QDocumentLineHandle *l)
{
	Block *b = m_blocks.count() ? m_blocks.last() : 0;
	
	if ( !b || (b->size() >= m_optimalSize) )
	{
		b = new Block(m_count);
		b->lines.reserve(m_optimalSize);
		m_blocks << b;
	}
	
	++m_count;
	b->lines.append(l);
	++b->end;
}

void QDocumentBuffer::insertLine(int index, QDocumentLineHandle *l)
{
	int blockIndex = blockForLine(index);
	
	if ( blockIndex == -1 )
	{
		appendLine(l);
		//qWarning("cannot insert line at pos %i", index);
		return;
	}
	
	++m_count;
	
	Block *b = m_blocks.at(blockIndex);
	
	if ( (b->size() + 1) >= m_forkThresold )
	{
		// split block
		int bounds = b->start + m_optimalSize;
		
		Block *nb = new Block(bounds);
		
		nb->insert(bounds, b->lines.constData() + m_optimalSize, b->size() - m_optimalSize);
		nb->lines.append(l);
		
		m_blocks.insert(blockIndex + 1, nb);
		b->lines.resize(m_optimalSize);
	} else {
		b->insert(index, l);
	}
	
	fixBlocksFrom(blockIndex);
}

void QDocumentBuffer::removeLine(int index)
{
	int blockIndex = blockForLine(index);
	
	if ( blockIndex == -1 )
	{
		qWarning("cannot remove line at pos %i", index);
		return;
	}
	
	--m_count;
	
	Block *b = m_blocks.at(blockIndex);
	
	b->remove(index);
	
	// if block too small, merge it with blocks around?
	
	if ( !b->size() )
	{
		// remove empty block
		m_blocks.remove(blockIndex);
		
		delete b;
	} else if ( 0 ) { //if ( b->size() < m_mergeThresold ) {
		// "merge" block with its neighbors
		
		int n = b->size();
		n += qMin(1, n / m_safetyRoom);
		
		int roomPrev = blockIndex ? m_forkThresold - m_blocks.at(blockIndex - 1)->size() : 0;
		int roomNext = blockIndex + 1 < m_blocks.count() ? m_forkThresold - m_blocks.at(blockIndex + 1)->size() : 0;
		
		bool maxPrev = false;
		int maxRoom = 0, minRoom = 0;
		Block *moreRoom = 0, *lessRoom = 0;
		
		if ( roomPrev > roomNext )
		{
			maxPrev = true;
			maxRoom = roomPrev;
			moreRoom = m_blocks.at(blockIndex - 1);
			
			minRoom = roomNext;
			
			if ( roomNext )
				lessRoom = m_blocks.at(blockIndex + 1);
		} else {
			maxRoom = roomNext;
			
			if ( roomNext )
				moreRoom = m_blocks.at(blockIndex + 1);
			
			minRoom = roomPrev;
			
			if ( roomPrev )
				lessRoom = m_blocks.at(blockIndex - 1);
		}
		
		if ( maxRoom > n )
		{
			// put everything in one
			moreRoom->lines << b->lines;
			moreRoom->end += b->size();
			
			m_blocks.remove(blockIndex);
			--blockIndex;
			delete b;
		} else if ( (maxRoom + minRoom) > n ) {
			// try to alloc evenly
			int part = b->size() * maxRoom / (maxRoom + minRoom);
			
			if ( maxPrev )
			{
				moreRoom->append(b->lines.constData(), part);
				lessRoom->prepend(b->lines.constData() + part, b->size() - part);
			} else {
				moreRoom->prepend(b->lines.constData(), part);
				lessRoom->append(b->lines.constData() + part, b->size() - part);
			}
			
			m_blocks.remove(blockIndex);
			--blockIndex;
			delete b;
		}
	}
	
	if ( m_lastBlock >= m_blocks.count() )
		m_lastBlock = -1;
	
	fixBlocksFrom(blockIndex);
}

void QDocumentBuffer::insertLines(int index, const QVector<QDocumentLineHandle*>& l)
{
	int blockIndex = blockForLine(index);
	
	if ( blockIndex == -1 )
	{
		foreach ( QDocumentLineHandle *h, l )
			appendLine(h);
		
		//qWarning("cannot insert line at pos %i", index);
		return;
	} else {
		//qDebug("inserting %i lines at %i (in block %i)", l.count(), index, blockIndex);
	}
	
	int n = l.count();
	
	m_count += n;
	
	Block *b = m_blocks.at(blockIndex);
	
	if ( (b->size() + n) >= m_forkThresold )
	{
		// split block
		int lidx = 0, max = n, bIdx = blockIndex;
		int bounds = b->start + m_optimalSize;
		
		Block *nb = new Block(bounds);
		
		if ( b->size() > m_optimalSize )
		{
			nb->insert(bounds, b->lines.constData() + m_optimalSize, b->size() - m_optimalSize);
			nb->append(l.constData(), n);
			m_blocks.insert(++bIdx, nb);
			b->lines.resize(m_optimalSize);
		} else {
			lidx = m_optimalSize - b->size();
			b->append(l.constData(), lidx);
			
			int chunk = qMin(m_optimalSize, max - lidx);
			nb->append(l.constData() + lidx, chunk);
			lidx += chunk;
		}
		
		while ( lidx < max )
		{
			int chunk = qMin(m_optimalSize, max - lidx);
			// do not specify start, will be fixed anyway
			nb = new Block(0);
			nb->append(l.constData() + lidx, chunk);
			m_blocks.insert(++bIdx, nb);
			lidx += chunk;
		}
		
	} else {
		b->insert(index, l.constData(), n);
	}
	
	fixBlocksFrom(blockIndex);
}

void QDocumentBuffer::removeLines(int index, int n)
{
	int blockIndex = blockForLine(index);
	
	if ( blockIndex == -1 )
	{
		qWarning("cannot remove line at pos %i", index);
		return;
	}
	
	m_count -= n;
	
	int count = n;
	int bIdx = blockIndex;
	Block *b = m_blocks.at(blockIndex);
	
	while ( count > 0 )
	{
		int room = b->end - index;
		int toRem = qMin(room, count);
		
		b->remove(index, toRem);
		count -= toRem;
		
		if ( !b->size() )
		{
			m_blocks.remove(bIdx);
			delete b;
		} else {
			++bIdx;
		}
		
		if ( bIdx >= m_blocks.count() )
			break;
		
		b = m_blocks.at(bIdx);
		index = b->start;
	}
	
	if ( count )
	{
		qWarning("Troubles in line removal");
	} else {
		//qDebug("%i blocks left : (%i, %i) : %i", m_blocks.count(), m_blocks.first()->start, m_blocks.first()->size(), m_count);
	}
	
	if ( false ) //b->size() < m_mergeThresold )
	{
		// "merge" block with its neighbors
		
		int sz = b->size();
		sz += qMin(1, sz / m_safetyRoom);
		
		int roomPrev = blockIndex ? m_forkThresold - m_blocks.at(blockIndex - 1)->size() : 0;
		int roomNext = blockIndex + 1 < m_blocks.count() ? m_forkThresold - m_blocks.at(blockIndex + 1)->size() : 0;
		
		bool maxPrev = false;
		int maxRoom = 0, minRoom = 0;
		Block *moreRoom = 0, *lessRoom = 0;
		
		if ( roomPrev > roomNext )
		{
			maxPrev = true;
			maxRoom = roomPrev;
			moreRoom = m_blocks.at(blockIndex - 1);
			
			minRoom = roomNext;
			
			if ( roomNext )
				lessRoom = m_blocks.at(blockIndex + 1);
		} else {
			maxRoom = roomNext;
			
			if ( roomNext )
				moreRoom = m_blocks.at(blockIndex + 1);
			
			minRoom = roomPrev;
			
			if ( roomPrev )
				lessRoom = m_blocks.at(blockIndex - 1);
		}
		
		if ( maxRoom > sz )
		{
			// put everything in one
			moreRoom->lines << b->lines;
			
			m_blocks.remove(blockIndex);
			--blockIndex;
			delete b;
		} else if ( (maxRoom + minRoom) > sz ) {
			// try to alloc evenly
			int part = b->size() * maxRoom / (maxRoom + minRoom);
			
			moreRoom->append(b->lines.constData(), part);
			lessRoom->append(b->lines.constData() + part, b->size() - part);
			
			m_blocks.remove(blockIndex);
			--blockIndex;
			delete b;
		}
	}
	
	if ( m_lastBlock >= m_blocks.count() )
		m_lastBlock = -1;
	
	fixBlocksFrom(blockIndex);
}

//////////////////////////////////////////////////////////////////////////////////////////

QDocumentBuffer::iterator::iterator(const iterator& i)
 : m_block(i.m_block), m_line(i.m_line), m_index(i.m_index), m_buffer(i.m_buffer), m_data(i.m_data), m_blockData(i.m_blockData)
{
	
}

QDocumentBuffer::iterator& QDocumentBuffer::iterator::operator = (const iterator& i)
{
	m_block = i.m_block;
	m_line = i.m_line;
	m_index = i.m_index;
	m_buffer = i.m_buffer;
	m_data = i.m_data;
	m_blockData = i.m_blockData;
	return *this;
}

bool QDocumentBuffer::iterator::atEnd() const
{
	return m_block == -1 || (m_buffer ? (m_line >= m_buffer->count()) : true);
}

int QDocumentBuffer::iterator::lineNumber() const
{
	return m_line;
}

QDocumentLineHandle* QDocumentBuffer::iterator::lineHandle() const
{
	//qDebug("line handle for %i : (%i, %i)", m_line, m_block, m_index);
	// no checks here : assume the user is smart enough to check for validity if using iterators
	return m_data[m_index];
	
	/*
	// TODO : make sure block coords are up-to-date?
	if ( m_block >= m_buffer->m_blocks.count() )
		return 0;
	
	Block *b = m_buffer->m_blocks.at(m_block);
	
	return m_index < b->lines.count() ? b->lines.at(m_index) : 0;
	*/
}

void QDocumentBuffer::iterator::moveTo(int lineNumber)
{
	m_line = lineNumber;
	
	if ( m_block != -1 && qAbs(m_line - lineNumber) < m_buffer->m_optimalSize )
	{
		// be smart
		Block *b = m_buffer->m_blocks.at(m_block);
		
		if ( b->start > lineNumber )
		{
			// try prev
			if ( m_block )
			{
				// TODO : add fine-grained blockForLine() with custom low/high
			} else {
				m_block = -1;
				m_index = -1;
			}
		} else if ( b->end <= lineNumber ) {
			// try next
			if ( m_block + 1 < m_buffer->m_blocks.count() )
			{
				// TODO : add fine-grained blockForLine() with custom low/high
			} else {
				m_block = -1;
				m_index = -1;
			}
		} else {
			// no need to move block :D
			m_index = lineNumber - b->start;
			return;
		}
	}
	
	m_block = m_buffer->blockForLine(lineNumber);
	
	if ( m_block != -1 )
	{
		m_blockData = m_buffer->m_blocks.at(m_block);
		m_data = m_blockData->lines.constData();
		m_index = lineNumber - m_blockData->start;
	} else {
		m_blockData = 0;
		m_data = 0;
	}
}

void QDocumentBuffer::iterator::moveBy(int numLines)
{
	m_line += numLines;
	
	m_index += numLines;
	
	if ( m_index < 0 )
	{
		// move backward
		
		while ( m_block > 0 && m_index < 0 )
		{
			m_blockData = m_buffer->m_blocks.at(--m_block);
			m_data = m_blockData->lines.constData();
			m_index += m_blockData->lines.count();
		}
	} else if ( m_line >= m_blockData->end ) {
		// move forward
		const int blockCount = m_buffer->m_blocks.count();
		int lineCount = m_blockData->lines.count();
		
		while ( (m_block + 1 < blockCount) && (m_index >= lineCount) )
		{
			m_index -= lineCount;
			m_blockData = m_buffer->m_blocks.at(++m_block);
			m_data = m_blockData->lines.constData();
			lineCount = m_blockData->lines.count();
		}
	} else {
		return;
	}
	
	// TODO : determine whether block using block boundaries would be better
	/*
	while ( m_block > 0 && m_index < 0 )
	{
		--m_block;
		m_index += m_buffer->m_blocks.at(m_block)->lines.count();
	}
	
	const int blockCount = m_buffer->m_blocks.count();
	int lineCount = m_buffer->m_blocks.at(m_block)->lines.count();
	
	while ( (m_block < blockCount) && (m_index >= lineCount) )
	{
		m_index -= lineCount;
		++m_block;
		
		if ( m_block < blockCount )
			lineCount = m_buffer->m_blocks.at(m_block)->lines.count();
	}
	*/
}

QDocumentBuffer::iterator::iterator(const QDocumentBuffer *buffer, int block, int line)
 : m_block(block), m_line(line), m_buffer(buffer), m_data(0), m_blockData(0)
{
	if ( m_block >= 0 && m_block < m_buffer->m_blocks.count() )
	{
		m_blockData = m_buffer->m_blocks.at(m_block);
		m_index = line - m_blockData->start;
		m_data = m_blockData->lines.constData();
	} else {
		m_index = 0;
		m_data = 0;
		m_blockData = 0;
	}
}
