/****************************************************************************
**
** Copyright (C) 2006-2009 fullmetalcoder <fullmetalcoder@hotmail.fr>
**
** This file is part of the Edyuk project <http://edyuk.org>
** 
** This file may be used under the terms of the GNU General Public License
** version 3 as published by the Free Software Foundation and appearing in the
** file GPL.txt included in the packaging of this file.
**
** This file is provided AS IS with NO WARRANTY OF ANY KIND, INCLUDING THE
** WARRANTY OF DESIGN, MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE.
**
****************************************************************************/

#include "qdocumentregiontree.h"

/*
	The purpose of this class is to abstract away of QDocument the management
	of hidden/wrapped areas
	
	The main idea is to split the document into regions to make updates (upon
	insertion/deletion) faster :
		* the data that used to be in QDocumentPrivate in unique maps for the
		whole document is now located inside the region it relates to
		* that data is stored using offsets relative to the region it resides
		in so upon updates only one region is modified
	
	The internal design is very similar to that of QDocumentBuffer but the purpose
	is completely different. The buffer holds the data of the document whereas the
	region tree (probably not the best name) holds render informations (which at
	some point should be moved from document to editor to allow proper sharing
	of document and things such as split views).
	
	TODO :
		* optimize the data structure ? (4 QMap<> takes a lot of spaces)
			-> 1 QVector<LineStruct*> for per-line "straightforward" storage (logical, span, QList<hidden>)
			-> better "cache" structure
		* improve the update algorithm?
		* offer option to dynamically turn caching on/off
*/

#include "qdocument_p.h"
#include "qdocumentbuffer.h"

QDocumentRegionTree::QDocumentRegionTree(QDocumentPrivate *d)
 : m_lastVisual(-1), m_lastLogical(-1)
{
	this->d = d;
}

int QDocumentRegionTree::visualLineCount() const
{
	if ( m_regions.isEmpty() )
		return 0;
	
	Region *r = m_regions.last();
	
	return r->visualStart + r->visualCount;
}

void QDocumentRegionTree::clearSpans(bool defer)
{
	foreach ( Region *r, m_regions )
	{
		r->m_wrapped.clear();
		
		if ( defer )
			r->dirty = true;
		else
			r->updateOffsets(0);
	}
	
	if ( defer )
		m_hasDeferred = true;
}

void QDocumentRegionTree::reset(int lineCount)
{
	qDeleteAll(m_regions);
	m_regions.clear();
	
	lineCount = qMax(1, lineCount);
	
	//qDebug("region tree init for %i lines", lineCount);
	
	int lastLogical = 0;
	
	while ( lineCount )
	{
		int sz = qMin((int)AverageRegionSize, lineCount);
		
		Region *r = new Region;
		r->logicalStart = lastLogical;
		r->logicalCount = sz;
		r->visualStart = lastLogical;
		r->visualCount = sz;
		m_regions << r;
		
		lastLogical += sz;
		lineCount -= sz;
	}
	
	//qDebug("-> %i regions", m_regions.count());
	
	m_firstDirtyRegion = m_regions.count();
	m_lastLogical = -1;
	m_lastVisual = -1;
}

void QDocumentRegionTree::linesRemoved(int logicalLine, int count, bool defer)
{
	// simple update : erase and shift
	
	int regionIndex = logicalRegion(logicalLine);
	
	if ( regionIndex == -1 )
		return;
	
	if ( m_firstDirtyRegion <= regionIndex )
	{
		fixRegionsFrom(m_firstDirtyRegion, regionIndex + 1);
	}
	
	int rIdx = regionIndex;
	Region *r = m_regions.at(rIdx);
	int offset = logicalLine - r->logicalStart;
	
	do
	{
		r = m_regions.at(rIdx);
		
		int chunk = qMin(count, r->logicalCount - offset);
		
		// TODO : merge when falling under MinimumRegionSize
		if ( chunk == r->logicalCount )
		{
			// deleting whole region...
			m_regions.remove(rIdx);
			count -= chunk;
			delete r;
			
			if ( !count )
				break;
			
			continue;
		}
		
		r->logicalCount -= chunk;
		
		// TODO : bypass old-fashion storage
		//r->m_hidden.insertMulti(offset, chunk);
		//r->updateOffsets(offset);
		
		QVector<int> smartBackup;
		QMap<int, int>::iterator it;
		const int sl = offset + chunk;
		
		// update hidden
		it = r->m_hidden.begin();
		
		while ( it != r->m_hidden.end() )
		{
			if ( it.key() >= offset )
				break;
			
			const int be = it.key() + *it;
			
			if ( be > offset )
			{
				// removing inside hidden block : shrink
				if ( be >= offset + count )
					*it -= count;
				else
					qWarning("unable to resolve hidden block/lines removal overlap");
				//hideLines = true;
			}
			
			++it;
		}
		
		//it = r->m_hidden.lowerBound(offset);
		
		while ( it != r->m_hidden.end() )
		{
			if ( it.key() + *it > sl )
			{
				//qDebug("mv(%i, %i)->(%i, %i)", it.key(), *it, it.key() - count, *it);
				smartBackup << it.key() - chunk << *it;
			}
			
			it = r->m_hidden.erase(it);
		}
		
		for ( int i = 0; i < smartBackup.count(); i += 2 )
			r->m_hidden[smartBackup.at(i)] = smartBackup.at(i + 1);
		
		smartBackup.clear();
		
		// update wrapped
		it = r->m_wrapped.lowerBound(offset);
		
		while ( it != r->m_wrapped.end() )
		{
			if ( it.key() + *it > sl )
			{
				//qDebug("mv(%i, %i)->(%i, %i)", it.key(), *it, it.key() - count, *it);
				smartBackup << it.key() - chunk << *it;
			}
			
			it = r->m_wrapped.erase(it);
		}
		
		for ( int i = 0; i < smartBackup.count(); i += 2 )
			r->m_wrapped[smartBackup.at(i)] = smartBackup.at(i + 1);
		
		if ( defer )
		{
			r->dirty = true;
			m_hasDeferred = true;
		} else
			r->updateOffsets(offset);
		
		count -= chunk;
		offset = 0;
		++rIdx;
	} while ( count );
	
	fixRegionsFrom(regionIndex);
}

void QDocumentRegionTree::linesInserted(int logicalLine, int count, bool defer)
{
	// complex update : shift and lookup
	// -> or simple update : shift and wait for setSpan() ?
	
	int regionIndex = logicalRegion(logicalLine);
	
	if ( regionIndex == -1 )
	{
		//qWarning("out of bounds insertion?");
		regionIndex = m_regions.count() - 1;
	}
	
	if ( m_firstDirtyRegion <= regionIndex )
	{
		fixRegionsFrom(m_firstDirtyRegion, regionIndex + 1);
	}
	
	int rIdx = regionIndex;
	Region *r = m_regions.at(rIdx);
	int offset = logicalLine - r->logicalStart;
	
	QDocumentBuffer::iterator dit = d->find(logicalLine);
	
	do
	{
		r = m_regions.at(rIdx);
		
		int chunk = qMin(r->logicalCount + count, (int)MaximumRegionSize) - r->logicalCount;
		
		if ( r->logicalCount + chunk >= MaximumRegionSize )
		{
			// TODO : split regions when they go beyong MaximumRegionSize
			
		}
		
		r->logicalCount += chunk;
		
		QVector<int> smartBackup;
		QMap<int, int>::iterator it;
		//const int sl = offset + chunk;
		
		bool hideLines = false;
		// update hidden
		it = r->m_hidden.begin();
		
		while ( it != r->m_hidden.end() )
		{
			if ( it.key() >= offset )
				break;
			
			if ( it.key() + *it > offset )
			{
				// inserting inside hidden block : expand
				*it += count;
				hideLines = true;
			}
			
			++it;
		}
		
		//it = r->m_hidden.lowerBound(offset);
		
		while ( it != r->m_hidden.end() )
		{
			smartBackup << it.key() + chunk << *it;
			
			it = r->m_hidden.erase(it);
		}
		
		for ( int i = 0; i < smartBackup.count(); i += 2 )
			r->m_hidden[smartBackup.at(i)] = smartBackup.at(i + 1);
		
		smartBackup.clear();
		
		// update wrapped
		it = r->m_wrapped.lowerBound(offset);
		
		while ( it != r->m_wrapped.end() )
		{
			smartBackup << it.key() + chunk << *it;
			
			it = r->m_wrapped.erase(it);
		}
		
		for ( int i = 0; i < smartBackup.count(); i += 2 )
			r->m_wrapped[smartBackup.at(i)] = smartBackup.at(i + 1);
		
		for ( int i = 0; i < chunk; ++i )
		{
			QDocumentLineHandle *h = dit.lineHandle();
			
			QVector<int> wb = d->wrapBoundaries(h);
			
			if ( wb.count() )
			{
				r->m_wrapped[offset + i] = wb.count();
			}
			
			dit.moveBy(1);
		}
		
		if ( defer )
		{
			r->dirty = true;
			m_hasDeferred = true;
		} else
			r->updateOffsets(offset);
		
		count -= chunk;
		offset = 0;
		++rIdx;
	} while ( count && rIdx < m_regions.count() );
	
	while ( count )
	{
		int sz = qMin((int)AverageRegionSize, count);
		
		Region *r = new Region;
		r->logicalCount = sz;
		r->visualCount = sz;
		m_regions << r;
		
		for ( int i = 0; i < sz; ++i )
		{
			QDocumentLineHandle *h = dit.lineHandle();
			
			QVector<int> wb = d->wrapBoundaries(h);
			
			if ( wb.count() )
			{
				r->m_wrapped[i] = wb.count();
			}
			
			dit.moveBy(1);
		}
		
		if ( defer )
		{
			r->dirty = true;
			m_hasDeferred = true;
		} else
			r->updateOffsets(offset);
		
		count -= sz;
	}
	
	fixRegionsFrom(regionIndex);
}

void QDocumentRegionTree::performDeferredUpdates() const
{
	int previousVisualBoundary = 0;
	int previousLogicalBoundary = 0;
	
	foreach ( Region *r, m_regions )
	{
		if ( r->dirty )
		{
			// TODO : store a dirty offset
			r->updateOffsets(0);
			r->dirty = false;
		}
		
		r->visualStart = previousVisualBoundary;
		r->logicalStart = previousLogicalBoundary;
		
		previousVisualBoundary += r->visualCount;
		previousLogicalBoundary += r->logicalCount;
	}
	
	m_hasDeferred = false;
}

void QDocumentRegionTree::Region::updateOffsets(int from)
{
	// TODO : be smarter, use a better data strucure
	// it should be possible to directly shift the offsets in-place without
	// reconstructing the whole offset data everytime something changes...
	//qDebug("updating offsets from %i", from);
	
	int offset = 0;
	QMap<int, int>::const_iterator hit, wit;
	const QMap<int, int>::const_iterator he = m_hidden.constEnd(), we = m_wrapped.constEnd();
	
	int visualFrom = visualLine(from);
	
	QMap<int, int>::iterator it;
	
	it = m_offsetsVT.lowerBound(visualFrom);
	
	while ( it != m_offsetsVT.end() )
	{
		//qDebug("erasing VT (%i, %i)", it.key(), *it);
		it = m_offsetsVT.erase(it);
	}
	
	it = m_offsetsTV.lowerBound(from);
	
	if ( it != m_offsetsTV.begin() )
	{
		offset = *(it -1) - (it -1).key();
		
		//qDebug("pre-offset : %i", offset);
	}
	
	while ( it != m_offsetsTV.end() )
	{
		//qDebug("erasing TV (%i, %i)", it.key(), *it);
		it = m_offsetsTV.erase(it);
	}
	
	//m_offsetsVT.clear();
	//m_offsetsTV.clear();
	
	hit = m_hidden.isEmpty() ? m_hidden.constBegin() : m_hidden.lowerBound(from);
	wit = m_wrapped.isEmpty() ? m_wrapped.constBegin() : m_wrapped.lowerBound(from);
	
	forever
	{
		const bool hiddenLeft = hit != he;
		const bool wrappedLeft = wit != we;
		
		if ( !(hiddenLeft || wrappedLeft) )
			break;
		
		const int hl = hiddenLeft ? hit.key() : logicalCount;
		const int wl = wrappedLeft ? wit.key() : logicalCount;
		const int local = qMin(hl, wl);
		
		int localOffset = 0;
		
		if ( wl <= hl )
		{
			// handle wrapped first
			localOffset = *wit;
			++wit;
			
			//m_offsetsVT[local + offset] = local;
			//offset += localOffset;
			//m_offsetsTV[local] = local + offset;
			
			m_offsetsVT[local + offset] = local;
			offset += localOffset;
			m_offsetsVT[local + offset] = local;
			m_offsetsTV[local] = local + offset;
		}
		
		if ( hl <= wl )
		{
			// handle hidden afterwards
			localOffset = *hit;
			
			forever
			{
				++hit;
				
				if ( hit == m_hidden.constEnd() || (hit.key() + *hit) > hl + localOffset )
					break;
				
				localOffset = qMax(hl + localOffset, hit.key() + *hit) - hl;
			}
			
			//qDebug("hidden(%i, %i)", hl, *hit);
			const int localEnd = local + localOffset;
			
			while ( wit != we && wit.key() <= localEnd )
			{
				++wit;
			}
			
			//qDebug("v:%i -> l:%i", local + offset, local + localOffset);
			//m_offsetsVT[local + offset + 1] = local + localOffset;
			//qDebug("l:%i -> v:%i", local + localOffset, local + offset);
			//m_offsetsTV[local + localOffset] = local + offset;
			//offset -= localOffset;
			
			m_offsetsVT[local + offset + 1] = local + localOffset + 1;
			m_offsetsTV[local] = local + offset;
			m_offsetsTV[local + localOffset] = local + offset;
			offset -= localOffset;
		}
	}
	
	if ( visualCount != logicalCount + offset )
	{
		visualCount = logicalCount + offset;
	}
}

int QDocumentRegionTree::Region::logicalLine(int visualLine, int *wrap) const
{
	QMap<int, int>::const_iterator
		cend = m_offsetsVT.constEnd(),
		cbeg = m_offsetsVT.constBegin(),
		it = m_offsetsVT.lowerBound(visualLine);
	
	if ( it == cend )
	{
		if ( it == cbeg )
			return visualLine;
		
		--it;
	}
	
	if ( it.key() == visualLine )
	{
		int logical = *it;
		
		if ( wrap && it != cbeg )
		{
			--it;
			
			if ( *it == logical )
				*wrap = visualLine - it.key();
		}
		return logical;
	} else if ( it.key() > visualLine ) {
		int next = *it;
		
		if ( it == cbeg )
			return visualLine;
		
		--it;
		
		if ( next == *it )
		{
			// within wrapped block
			
			if ( wrap )
				*wrap = visualLine - it.key();
			
			return next;
		}
	}
	
	const int vl = it.key();
	
	int tl = *it;
	int voff = visualLine - vl;
	
	tl += voff;
	//qDebug("visual:%i @ %i->%i, %i", visualLine, vl, tl, voff);
	
	/*
	if ( voff )
	{
		// wrap bounds checking
		//QDocumentLineHandle *h = m_buffer->at(tl);
		//QVector<int> bounds = wrapBoundaries(h);
		
		const QMap<int, int>::const_iterator it = m_wrapped.find(tl);
		
		if ( it != m_wrapped.end() )
		{
			// not on the right line...
			
			if ( voff > *it )
			{
				voff -= *it;
				tl += voff;
				voff = 0;
			}
		} else {
			tl += voff;
			voff = 0;
		}
	}
	
	if ( wrap )
	{
		//qDebug("wrap : %i", voff);
		*wrap = voff;
	}
	*/
	
	//qDebug("v:%i>%i", visualLine, tl);
	
	return tl;
}

int QDocumentRegionTree::Region::visualLine(int logicalLine) const
{
	QMap<int, int>::const_iterator
		cend = m_offsetsTV.constEnd(),
		cbeg = m_offsetsTV.constBegin(),
		it = m_offsetsTV.lowerBound(logicalLine);
	
	if ( it == m_offsetsTV.end() )
	{
		if ( it == m_offsetsTV.begin() )
			return logicalLine;
		
		--it;
	}
	
	if ( it.key() == logicalLine )
	{
		if ( it == cbeg )
			return logicalLine;
		
		--it;
		
		//qDebug("how come");
		//return *it;
	} else if ( it.key() > logicalLine ) {
		int next = *it;
		
		if ( it == cbeg )
			return logicalLine;
		
		--it;
		
		if ( next == *it )
		{
			// within wrapped block
			return next;
		}
	}
	
	const int tl = it.key();
	const int vl = *it;
	
	int toff = logicalLine - tl;
	
	//qDebug("l:%i>%i", logicalLine, vl + toff);
	
	return vl + toff;
}

void QDocumentRegionTree::hide(int logicalLine, int count, bool defer)
{
	int regionIndex = logicalRegion(logicalLine);
	
	if ( regionIndex == -1 )
		return;
	
	if ( m_firstDirtyRegion <= regionIndex )
	{
		fixRegionsFrom(m_firstDirtyRegion, regionIndex + 1);
	}
	
	int rIdx = regionIndex;
	Region *r = m_regions.at(rIdx);
	int offset = logicalLine - r->logicalStart;
	
	do
	{
		r = m_regions.at(rIdx);
		
		int chunk = qMin(count, r->logicalCount);
		
		// TODO : bypass old-fashion storage
		r->m_hidden.insertMulti(offset, chunk);
		
		if ( defer )
		{
			r->dirty = true;
			m_hasDeferred = true;
		} else
			r->updateOffsets(offset);
		
		count -= chunk;
		offset = 0;
		++rIdx;
	} while ( count );
	
	fixRegionsFrom(regionIndex);
}

void QDocumentRegionTree::show(int logicalLine, int count, bool defer)
{
	int regionIndex = logicalRegion(logicalLine);
	
	if ( regionIndex == -1 )
		return;
	
	if ( m_firstDirtyRegion <= regionIndex )
	{
		fixRegionsFrom(m_firstDirtyRegion, regionIndex + 1);
	}
	
	int rIdx = regionIndex;
	Region *r = m_regions.at(rIdx);
	int offset = logicalLine - r->logicalStart;
	
	do
	{
		r = m_regions.at(rIdx);
		
		int chunk = qMin(count, r->logicalCount);
		
		// TODO : bypass old-fashion storage
		
		QMap<int, int>::iterator it = r->m_hidden.find(offset);
		
		while ( (it != r->m_hidden.end()) && (it.key() == offset)  )
		{
			if ( *it == count )
			{
				//qDebug("showing %i lines from %i", count, line);
				it = r->m_hidden.erase(it);
			} else
				++it;
		}
		
		if ( defer )
		{
			r->dirty = true;
			m_hasDeferred = true;
		} else
			r->updateOffsets(offset);
		
		count -= chunk;
		offset = 0;
		++rIdx;
	} while ( count );
	
	fixRegionsFrom(regionIndex);
}

void QDocumentRegionTree::setSpan(int logicalLine, int count, bool defer)
{
	int regionIndex = logicalRegion(logicalLine);
	
	if ( regionIndex == -1 )
		return;
	
	Region *r = m_regions.at(regionIndex);
	
	if ( m_firstDirtyRegion <= regionIndex )
	{
		fixRegionsFrom(m_firstDirtyRegion, regionIndex + 1);
	}
	
	int offset = logicalLine - r->logicalStart;
	
	// TODO : bypass old-fashion storage
	QMap<int, int>::iterator it = r->m_wrapped.find(offset);
	
	//qDebug("line %i spans over %i", logicalLine, count);
	
	if ( it == r->m_wrapped.end() )
	{
		if ( count )
		{
			r->m_wrapped[offset] = count;
			r->visualCount += count;
		}
	} else {
		if ( *it != count )
		{
			r->visualCount += (count - *it);
			
			if ( count )
				*it = count;
			else
				r->m_wrapped.erase(it);
		} else {
			return;
		}
	}
	
	// TODO : smarter update in this case should be doable
	if ( defer )
	{
		r->dirty = true;
		m_hasDeferred = true;
		m_firstDirtyRegion = qMin(m_firstDirtyRegion, regionIndex);
	} else {
		r->updateOffsets(offset);
		fixRegionsFrom(regionIndex);
	}
}

void QDocumentRegionTree::fixRegionsFrom(int lastValidRegion, int maxRegion)
{
	// update region boundaries
	// this is quite fast as it only write two integers into each region,
	// involves no memory operation and operate on a small number of regions
	// (the region count should not go beyond a thousand, which would already
	// translate to a ~10 Mlines document...)
	
	if ( lastValidRegion < 0 )
		lastValidRegion = 0;
	if ( lastValidRegion >= m_regions.count() )
		return;
	
	if ( maxRegion < 0 || maxRegion > m_regions.count() )
		maxRegion = m_regions.count();
	
	Region *r = lastValidRegion ? m_regions.at(lastValidRegion - 1) : 0;
	int previousVisualBoundary = r ? r->visualStart + r->visualCount : 0;
	int previousLogicalBoundary = r ? r->logicalStart + r->logicalCount : 0;
	
	while ( lastValidRegion < maxRegion )
	{
		r = m_regions.at(lastValidRegion);
		
		r->visualStart = previousVisualBoundary;
		r->logicalStart = previousLogicalBoundary;
		
		previousVisualBoundary += r->visualCount;
		previousLogicalBoundary += r->logicalCount;
		
		++lastValidRegion;
	}
	
	m_firstDirtyRegion = maxRegion;
}

int QDocumentRegionTree::logicalLineNumber(int visualLine, int *wrap) const
{
	int regionIndex = visualRegion(visualLine);
	
	if ( m_hasDeferred )
	{
		performDeferredUpdates();
	}
	
	if ( regionIndex != -1 )
	{
		Region *r = m_regions.at(regionIndex);
		
		int ll = r->logicalStart + r->logicalLine(visualLine - r->visualStart, wrap);
		
		//qDebug("visual:%i -> logical:%i", visualLine, ll);
		
		return ll;
	}
	
	return -1;
}

int QDocumentRegionTree::visualLineNumber(int logicalLine) const
{
	int regionIndex = logicalRegion(logicalLine);
	
	if ( m_hasDeferred )
	{
		performDeferredUpdates();
	}
	
	if ( regionIndex != -1 )
	{
		Region *r = m_regions.at(regionIndex);
		
		int vl = r->visualStart + r->visualLine(logicalLine - r->logicalStart);
		
		//qDebug("logical:%i -> visual:%i", logicalLine, vl);
		
		return vl;
	}
	
	return -1;
}

int QDocumentRegionTree::visualRegion(int line) const
{
	if ( line < 0 || (line >= (m_regions.last()->visualStart + m_regions.last()->visualCount)) )
		return -1;
	
	Region *r = 0;
	int low = 0, high = m_regions.count() - 1;
	
	if ( m_lastVisual != -1 )
	{
		// tries to smartly improve perf for consecutive search
		r = m_regions.at(m_lastVisual);
		
		const int off = line - r->visualStart;
		
		if ( off < 0  )
			high = qMax(low, m_lastVisual - 1);
		else if ( off >= r->visualCount )
			low = qMin(high, m_lastVisual + 1);
		else
			return m_lastVisual;
		 
	}
	
	if ( line < (m_regions.at(low)->visualStart + m_regions.at(low)->visualCount) )
	{
		m_lastVisual = low;
		return low;
	}
	
	if ( line >= m_regions.at(high)->visualStart )
	{
		m_lastVisual = high;
		return high;
	}
	
	while ( low < high )
	{
		int mid = (low + high) / 2;
		r = m_regions.at(mid);
		
		const int off = line - r->visualStart;
		
		if ( off < 0  )
			high = mid;
		else if ( off >= r->visualCount )
			low = mid;
		else {
			m_lastVisual = mid;
			return mid;
		}
	}
	
	m_lastVisual = low;
	
	return low;
}

int QDocumentRegionTree::logicalRegion(int line) const
{
	if ( line < 0 || (line >= (m_regions.last()->logicalStart + m_regions.last()->logicalCount)) )
		return -1;
	
	Region *r = 0;
	int low = 0, high = m_regions.count() - 1;
	
	if ( m_lastLogical != -1 )
	{
		// tries to smartly improve perf for consecutive search
		r = m_regions.at(m_lastLogical);
		
		const int off = line - r->logicalStart;
		
		if ( off < 0 )
			high = qMax(low, m_lastLogical - 1);
		else if ( off >= r->logicalCount )
			low = qMin(high, m_lastLogical + 1);
		else
			return m_lastLogical;
		 
	}
	
	if ( line < (m_regions.at(low)->logicalStart + m_regions.at(low)->logicalCount) )
	{
		m_lastLogical = low;
		return low;
	}
	
	if ( line >= m_regions.at(high)->logicalStart )
	{
		m_lastLogical = high;
		return high;
	}
	
	while ( low < high )
	{
		int mid = (low + high) / 2;
		r = m_regions.at(mid);
		
		const int off = line - r->logicalStart;
		
		if ( off < 0  )
			high = mid;
		else if ( off >= r->logicalCount )
			low = mid;
		else {
			m_lastLogical = mid;
			return mid;
		}
	}
	
	m_lastLogical = low;
	
	return low;
}

